
{ Autor:   Jürgen Schlottke, Schönaich-C.-Str. 46, D-W 2200 Elmshorn
           Tel. 04121/63109
  Zweck  : Demonstration der Schachprogrammierung unter Turbo-Pascal
  Datum  : irgendwann 1991, als PD freigegeben am 18.01.93
  Version: ohne Versionsnummer
}

unit ChessPlayer;

interface

uses
  ChessPlayerCore;

type
  TExitCode = (ecSuccess, ecCheck, ecCheckmate, ecStalemate, ecError);
  
  TChessPlayer = class
  strict private
    FIniPos,
    FCurPos: TChessPosition;
    FHalfmovesCount: integer;
    FSecondList: TMoveList;
  public
    constructor Create;
    destructor Destroy; override;
    procedure SetPosition(const APos: string);
    function PlayMove(const AMove: integer): boolean;
    function BestMoveIndex(const ABestValue: integer): integer;
    function BestMove(out ACode: TExitCode): integer; overload;
    function BestMove: integer; overload;
    function BoardAsText(const APretty: boolean = TRUE): string;
    function FEN: string;
    //function SetPositionFromMoves(const APos: string; const AMoves: array of string): string;
  end;

implementation

uses
  Classes, SysUtils, TypInfo, ChessPlayerTypes, Log;

const
  STARTPOS = 'rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1';

constructor TChessPlayer.Create;
begin
  inherited Create;
  FIniPos := TChessPosition.Create(STARTPOS);
  FCurPos := TChessPosition.Create;
end;

destructor TChessPlayer.Destroy;
begin
  FCurPos.Free;
  FIniPos.Free;
  inherited Destroy;
end;

procedure TChessPlayer.SetPosition(const APos: string);

  function GetMoveCount(const APos: string): integer;
  var
    LFields: TStringList;
  begin
    LFields := TStringList.Create;
    ExtractStrings([' '], [], pchar(APos), LFields);
    if LFields.Count = 6 then
      result := 2 * Pred(StrToInt(LFields[5])) + Ord(LFields[1] = 'b')
    else
      result := 0;
    LFields.Free;
  end;

begin
  //Log(Format('%s SetPosition(%s)', [TimeToStr(Time), APos]));
  FCurPos.SetPosition(APos);
  FHalfmovesCount := GetMoveCount(APos);
end;

function TChessPlayer.PlayMove(const AMove: integer): boolean;
var
  LPos: TChessPosition;
  LMove, LPromo, LFrom, LTo: integer;
begin
  ToLog({$I %FILE%}, {$I %LINE%}, Format('PlayMove(%d) [FCurPos.Active = %d]', [AMove, FCurPos.Active]));
  LMove := AMove;
  LPromo := LMove div 10000;
  LFrom := (LMove mod 10000) div 100;
  LTo := LMove mod 100;
  FCurPos.GenerateMoves;
  result := FCurPos.IsLegal(LMove);
  if result then
  begin
    LPos := TChessPosition.Create;
    LPos.SetPosition(FCurPos);
    LPos.MovePiece(LFrom, LTo, LPromo);
    LPos.Active := CBlack * LPos.Active;
    if LPos.Check then
      result := FALSE
    else
    begin
      FCurPos.MovePiece(LFrom, LTo, LPromo);
      Inc(FHalfmovesCount);
    end;
    LPos.Free;
  end;
  ToLog({$I %FILE%}, {$I %LINE%}, Format('PlayMove = %s [FCurPos.Active = %d]', [BoolToStr(result, TRUE), FCurPos.Active]));
end;

function TChessPlayer.BestMoveIndex(const ABestValue: integer): integer;
var
  LPos: TChessPosition;
  LMaxValue: integer;
  LSecondListCount: integer;
  i, j: integer;
begin
  LPos := TChessPosition.Create;

  with FCurPos do
  begin
    LSecondListCount := 0;
    for i := 1 to MoveCount do
      if MoveList[i].FValue = ABestValue then
      begin
        Inc(LSecondListCount);
        FSecondList[LSecondListCount].FFrom := MoveList[i].FFrom;
        FSecondList[LSecondListCount].FTo := MoveList[i].FTo;
        FSecondList[LSecondListCount].FValue := 0
      end;
  end;

  LMaxValue := Low(integer);
  result := 0;
  
  for i := 1 to LSecondListCount do
  begin
    LPos.SetPosition(FCurPos);

    with LPos do
    begin
      if Board[FSecondList[i].FFrom] = FIniPos.Board[FSecondList[i].FFrom] then
      begin
        Inc(FSecondList[i].FValue, 5);
        if Board[FSecondList[i].FFrom] * Active = CPawn then
          Inc(FSecondList[i].FValue, 2);
      end;

      if Board[FSecondList[i].FFrom] * Active = CKing then
        if Abs(FSecondList[i].FTo - FSecondList[i].FFrom) = 20 then
          Inc(FSecondList[i].FValue, 20)
        else
          Dec(FSecondList[i].FValue, 10);

      if (FHalfmovesCount < 32) and (Board[FSecondList[i].FFrom] * Active in [CPawn, CBishop, CKnight]) then
        Inc(FSecondList[i].FValue, 20);

      if (FSecondList[i].FFrom div 10 = 1)
      or (FSecondList[i].FFrom div 10 = 8)
      or (FSecondList[i].FFrom mod 10 = 1)
      or (FSecondList[i].FFrom mod 10 = 8) then
        Inc(FSecondList[i].FValue, 2);

      if (FSecondList[i].FTo div 10 = 1)
      or (FSecondList[i].FTo div 10 = 8)
      or (FSecondList[i].FTo mod 10 = 1)
      or (FSecondList[i].FTo mod 10 = 8) then
        Dec(FSecondList[i].FValue, 2);
    end;

    LPos.SetPosition(FCurPos);
    LPos.MovePiece(FSecondList[i].FFrom, FSecondList[i].FTo, CQueen);

    if LPos.Board[FSecondList[i].FTo] = FIniPos.Board[FSecondList[i].FTo] then
      Dec(FSecondList[i].FValue, 10);

    LPos.Active := CBlack * LPos.Active;
    LPos.GenerateSimpleMoves;

    with LPos do
      for j := 1 to MoveCount do
      begin
        Inc(FSecondList[i].FValue);
        if Board[MoveList[j].FTo] <> CNil then
          Inc(FSecondList[i].FValue);
      end;

    LPos.Active := CBlack * LPos.Active;
    LPos.GenerateSimpleMoves;

    with LPos do
      for j := 1 to MoveCount do
      begin
        Dec(FSecondList[i].FValue);
        if Board[MoveList[j].FTo] <> CNil then
          Dec(FSecondList[i].FValue);
      end;
    if FSecondList[i].FValue >= LMaxValue then
    begin
      LMaxValue := FSecondList[i].FValue;
      result := i;
    end;
  end;
  LPos.Free;
end;

function TChessPlayer.BestMove(out ACode: TExitCode): integer;
const
  CCodeIllegal: array[boolean] of TExitCode = (ecStalemate, ecCheckmate);
  CCodeLegal: array[boolean] of TExitCode = (ecSuccess, ecCheck);
var
  i: integer;
  LCheckBefore, LPromo: boolean;
  LPos: TChessPosition;
begin
  result := {'}0000{'};
  i := FCurPos.BestEval(FCurPos.Active, 1, 32000);
  i := BestMoveIndex(i);
  if i > 0 then
  begin
    LPos := TChessPosition.Create;
    LPos.SetPosition(FCurPos);
    //Inc(FHalfmovesCount);
    //LCheckBefore := FCurPos.Check;
    LCheckBefore := LPos.Check;
    with FSecondList[i] do
      LPromo := {FCurPos.}LPos.MovePiece(FFrom, FTo, CQueen);
    {FCurPos.}LPos.Active := CBlack * {FCurPos.}LPos.Active;
    if {FCurPos.}LPos.Check then
      ACode := CCodeIllegal[LCheckBefore]
    else
    begin
      with FSecondList[i] do
        result := {MoveToStr(}100 * FFrom + FTo{)};
      (*
      if LPromo then
        result := result + 'q';
      *)
      {FCurPos.}LPos.Active := CBlack * {FCurPos.}LPos.Active;
      ACode := CCodeLegal[{FCurPos.}LPos.Check];
    end;
    LPos.Free;
  end else
    ACode := ecError;
  if not (ACode in [ecSuccess, ecCheck]) then
    //Log(Format('bestmove function exit code %d', [ACode]));
end;

function TChessPlayer.BestMove: integer;
var
  LCode: TExitCode;
begin
  //Log(Format('%s BestMove', [TimeToStr(Time)]));
  //Log(BoardAsText(TRUE));
  result := BestMove(LCode);
end;

function TChessPlayer.BoardAsText(const APretty: boolean): string;
begin
  result := FCurPos.BoardAsText(APretty);
end;

function TChessPlayer.FEN: string;
begin
  result := FCurPos.FEN;
end;

(*
function TChessPlayer.SetPositionFromMoves(const APos: string; const AMoves: array of string): string;
var
  i: integer;
begin
  SetPosition(APos);
  for i := Low(AMoves) to High(AMoves) do
    PlayMove(AMoves[i]);
  result := FEN;
end;
*)

end.
